from abc import ABC, abstractmethod
from enum import Enum
from typing import List

from pydantic import field_validator, BaseModel

from traveltimepy import Coordinates
from traveltimepy.wkt.helper import print_indented


class GeometryType(Enum):
    POINT = "Point"
    LINESTRING = "LineString"
    POLYGON = "Polygon"
    MULTIPOINT = "MultiPoint"
    MULTILINESTRING = "MultiLineString"
    MULTIPOLYGON = "MultiPolygon"


class WKTObject(BaseModel, ABC):
    type: GeometryType

    @abstractmethod
    def pretty_print(self, indent_level=0):
        raise NotImplementedError("Subclasses must implement this method.")


class PointModel(WKTObject):
    coordinates: Coordinates

    def __init__(self, **data):
        super().__init__(type=GeometryType.POINT, **data)

    def pretty_print(self, indent_level=0):
        print_indented(
            f"POINT: {self.coordinates.lat}, {self.coordinates.lng}", indent_level
        )


class LineStringModel(WKTObject):
    coordinates: List[PointModel]

    def __init__(self, **data):
        super().__init__(type=GeometryType.LINESTRING, **data)

    @field_validator("coordinates")
    @classmethod
    def check_minimum_coordinates(cls, coords):
        if len(coords) < 2:
            raise ValueError("LineString must have at least 2 coordinates.")
        return coords

    def pretty_print(self, indent_level=0):
        print_indented("LINE STRING:", indent_level)
        for point in self.coordinates:
            point.pretty_print(indent_level + 1)


class PolygonModel(WKTObject):
    exterior: LineStringModel
    interiors: List[LineStringModel]

    def __init__(self, **data):
        super().__init__(type=GeometryType.POLYGON, **data)

    def pretty_print(self, indent_level=0):
        print_indented("POLYGON:", indent_level)
        print_indented("EXTERIOR:", indent_level + 1)
        self.exterior.pretty_print(indent_level + 2)
        if self.interiors:
            print_indented("INTERIORS:", indent_level + 1)
            for interior in self.interiors:
                interior.pretty_print(indent_level + 2)


class MultiPointModel(WKTObject):
    coordinates: List[PointModel]

    def __init__(self, **data):
        super().__init__(type=GeometryType.MULTIPOINT, **data)

    def pretty_print(self, indent_level=0):
        print_indented("MULTIPOINT:", indent_level)
        for point in self.coordinates:
            point.pretty_print(indent_level + 1)


class MultiLineStringModel(WKTObject):
    coordinates: List[LineStringModel]

    def __init__(self, **data):
        super().__init__(type=GeometryType.MULTILINESTRING, **data)

    def pretty_print(self, indent_level=0):
        print_indented("MULTILINESTRING:", indent_level)
        for linestring in self.coordinates:
            linestring.pretty_print(indent_level + 1)


class MultiPolygonModel(WKTObject):
    coordinates: List[PolygonModel]

    def __init__(self, **data):
        super().__init__(type=GeometryType.MULTIPOLYGON, **data)

    def pretty_print(self, indent_level=0):
        print_indented("MULTIPOLYGON:", indent_level)
        for polygon in self.coordinates:
            polygon.pretty_print(indent_level + 1)



from datetime import datetime
from typing import List, Optional, Union

from pydantic.main import BaseModel

from traveltimepy.dto.common import Location, Property, FullRange, Snapping
from traveltimepy.dto.transportation import (
    PublicTransport,
    Driving,
    Ferry,
    Walking,
    Cycling,
    DrivingTrain,
    CyclingPublicTransport,
)
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.responses.routes import RoutesResponse
from traveltimepy.itertools import split, flatten


class ArrivalSearch(BaseModel):
    id: str
    departure_location_ids: List[str]
    arrival_location_id: str
    arrival_time: datetime
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ]
    properties: List[Property]
    range: Optional[FullRange] = None
    snapping: Optional[Snapping] = None


class DepartureSearch(BaseModel):
    id: str
    arrival_location_ids: List[str]
    departure_location_id: str
    departure_time: datetime
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ]
    properties: List[Property]
    range: Optional[FullRange] = None
    snapping: Optional[Snapping] = None


class RoutesRequest(TravelTimeRequest[RoutesResponse]):
    locations: List[Location]
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]

    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        return [
            RoutesRequest(
                locations=self.locations,
                departure_searches=departures,
                arrival_searches=arrivals,
            )
            for departures, arrivals in split(
                self.departure_searches, self.arrival_searches, window_size
            )
        ]

    def merge(self, responses: List[RoutesResponse]) -> RoutesResponse:
        return RoutesResponse(
            results=flatten([response.results for response in responses])
        )



import os
from typing import List

import pytest

from traveltimepy import Location, Coordinates
from traveltimepy.sdk import TravelTimeSdk


@pytest.fixture
def sdk() -> TravelTimeSdk:
    return TravelTimeSdk(os.environ["APP_ID"], os.environ["API_KEY"])


@pytest.fixture
def proto_sdk() -> TravelTimeSdk:
    return TravelTimeSdk(os.environ["PROTO_APP_ID"], os.environ["PROTO_API_KEY"])


@pytest.fixture
def locations() -> List[Location]:
    return [
        Location(id="London center", coords=Coordinates(lat=51.508930, lng=-0.131387)),
        Location(id="Hyde Park", coords=Coordinates(lat=51.508824, lng=-0.167093)),
        Location(id="ZSL London Zoo", coords=Coordinates(lat=51.536067, lng=-0.153596)),
    ]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, TypeVar, Generic

from pydantic import BaseModel

T = TypeVar("T")


class TravelTimeRequest(ABC, BaseModel, Generic[T]):
    @abstractmethod
    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        pass

    @abstractmethod
    def merge(self, responses: List[T]) -> T:
        pass



import pytest  # noqa

from traveltimepy import Coordinates
from traveltimepy.wkt import (
    parse_wkt,
    PointModel,
    LineStringModel,
    PolygonModel,
    MultiPointModel,
    MultiLineStringModel,
    MultiPolygonModel,
)
from traveltimepy.wkt.error import (
    InvalidWKTStringError,
    NullGeometryError,
    InvalidGeometryTypeError,
)

point_wkt = "POINT (0 0)"
line_wkt = "LINESTRING(0 0, 1 1, 2 2)"
poly_wkt = "POLYGON((0 0, 0 2, 2 2, 2 0, 0 0))"
mp_wkt = "MULTIPOINT(0 0, 1 1)"
mls_wkt = "MULTILINESTRING((0 0, 1 1), (2 2, 3 3))"
mpoly_wkt = "MULTIPOLYGON(((0 0, 0 2, 2 2, 2 0, 0 0)))"


def test_parse_point():
    parsed = parse_wkt(point_wkt)
    assert parsed == PointModel(coordinates=Coordinates(lat=0, lng=0))


def test_parse_line_string():
    parsed = parse_wkt(line_wkt)
    assert parsed == LineStringModel(
        coordinates=[
            PointModel(coordinates=Coordinates(lat=0.0, lng=0.0)),
            PointModel(coordinates=Coordinates(lat=1.0, lng=1.0)),
            PointModel(coordinates=Coordinates(lat=2.0, lng=2.0)),
        ],
    )


def test_parse_polygon():
    parsed = parse_wkt(poly_wkt)
    assert parsed == PolygonModel(
        exterior=LineStringModel(
            coordinates=[
                PointModel(coordinates=Coordinates(lat=0.0, lng=0.0)),
                PointModel(coordinates=Coordinates(lat=0.0, lng=2.0)),
                PointModel(coordinates=Coordinates(lat=2.0, lng=2.0)),
                PointModel(coordinates=Coordinates(lat=2.0, lng=0.0)),
                PointModel(coordinates=Coordinates(lat=0.0, lng=0.0)),
            ],
        ),
        interiors=[],
    )


def test_parse_multi_point():
    parsed = parse_wkt(mp_wkt)
    assert parsed == MultiPointModel(
        coordinates=[
            PointModel(coordinates=Coordinates(lat=0.0, lng=0.0)),
            PointModel(coordinates=Coordinates(lat=1.0, lng=1.0)),
        ],
    )


def test_parse_multi_line_string():
    parsed = parse_wkt(mls_wkt)
    assert parsed == MultiLineStringModel(
        coordinates=[
            LineStringModel(
                coordinates=[
                    PointModel(
                        coordinates=Coordinates(lat=0.0, lng=0.0),
                    ),
                    PointModel(
                        coordinates=Coordinates(lat=1.0, lng=1.0),
                    ),
                ],
            ),
            LineStringModel(
                coordinates=[
                    PointModel(
                        coordinates=Coordinates(lat=2.0, lng=2.0),
                    ),
                    PointModel(
                        coordinates=Coordinates(lat=3.0, lng=3.0),
                    ),
                ],
            ),
        ],
    )


def test_parse_multi_polygon():
    parsed = parse_wkt(mpoly_wkt)
    assert parsed == MultiPolygonModel(
        coordinates=[
            PolygonModel(
                exterior=LineStringModel(
                    coordinates=[
                        PointModel(
                            coordinates=Coordinates(lat=0.0, lng=0.0),
                        ),
                        PointModel(
                            coordinates=Coordinates(lat=0.0, lng=2.0),
                        ),
                        PointModel(
                            coordinates=Coordinates(lat=2.0, lng=2.0),
                        ),
                        PointModel(
                            coordinates=Coordinates(lat=2.0, lng=0.0),
                        ),
                        PointModel(
                            coordinates=Coordinates(lat=0.0, lng=0.0),
                        ),
                    ],
                ),
                interiors=[],
            )
        ],
    )


def test_invalid_wkt_string():
    with pytest.raises(InvalidWKTStringError):
        parse_wkt("INVALIDWKTSTRING")


def test_null_geometry():
    with pytest.raises(NullGeometryError):
        parse_wkt("POINT EMPTY")


def test_unsupported_geometry_type():
    unsupported_wkt = "GEOMETRYCOLLECTION(POINT(2 3),LINESTRING(2 3, 3 4))"
    with pytest.raises(InvalidGeometryTypeError):
        parse_wkt(unsupported_wkt)



import pytest

from traveltimepy import Transportation, TravelTimeSdk


@pytest.mark.asyncio
async def test_one_to_many(sdk: TravelTimeSdk, locations):
    results = await sdk.time_filter_fast_async(
        locations=locations,
        search_ids={
            "London center": ["Hyde Park", "ZSL London Zoo"],
            "ZSL London Zoo": ["Hyde Park", "London center"],
        },
        transportation=Transportation(type="public_transport"),
    )

    assert len(results) > 0


@pytest.mark.asyncio
async def test_many_to_one(sdk: TravelTimeSdk, locations):
    results = await sdk.time_filter_fast_async(
        locations=locations,
        search_ids={
            "London center": ["Hyde Park", "ZSL London Zoo"],
            "ZSL London Zoo": ["Hyde Park", "London center"],
        },
        transportation=Transportation(type="public_transport"),
        one_to_many=False,
    )

    assert len(results) > 0



import random
from typing import List

from traveltimepy import Location, Coordinates


def generate_float(value: float, radius: float) -> float:
    return random.uniform(value - radius, value + radius)


def generate_locations(
    lat: float, lng: float, radius: float, name: str, amount: int
) -> List[Location]:
    return [
        Location(
            id="{} {}".format(name, i),
            coords=Coordinates(
                lat=generate_float(lat, radius), lng=generate_float(lng, radius)
            ),
        )
        for i in range(amount)
    ]


def generate_coordinates(
    lat: float, lng: float, radius: float, amount: int
) -> List[Coordinates]:
    return [
        Coordinates(lat=generate_float(lat, radius), lng=generate_float(lng, radius))
        for _ in range(amount)
    ]



import pytest

from traveltimepy import Transportation, TravelTimeSdk
from traveltimepy.dto.common import Coordinates


@pytest.mark.asyncio
async def test_one_to_many(sdk: TravelTimeSdk):
    results = await sdk.time_map_fast_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        travel_time=900,
        transportation=Transportation(type="public_transport"),
    )

    assert len(results) == 2


@pytest.mark.asyncio
async def test_many_to_one(sdk: TravelTimeSdk):
    results = await sdk.time_map_fast_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        travel_time=900,
        transportation=Transportation(type="public_transport"),
        one_to_many=False,
    )

    assert len(results) == 2


@pytest.mark.asyncio
async def test_one_to_many_geojson(sdk: TravelTimeSdk):
    results = await sdk.time_map_fast_geojson_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        travel_time=900,
        transportation=Transportation(type="public_transport"),
    )

    assert len(results) == 2


@pytest.mark.asyncio
async def test_many_to_one_geojson(sdk: TravelTimeSdk):
    results = await sdk.time_map_fast_geojson_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        travel_time=900,
        transportation=Transportation(type="public_transport"),
        one_to_many=False,
    )

    assert len(results) == 2


@pytest.mark.asyncio
async def test_one_to_many_wkt(sdk: TravelTimeSdk):
    response = await sdk.time_map_fast_wkt_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        travel_time=900,
        transportation=Transportation(type="public_transport"),
    )

    assert len(response.results) == 2


@pytest.mark.asyncio
async def test_many_to_one_wkt(sdk: TravelTimeSdk):
    response = await sdk.time_map_fast_wkt_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        travel_time=900,
        transportation=Transportation(type="public_transport"),
        one_to_many=False,
    )

    assert len(response.results) == 2


@pytest.mark.asyncio
async def test_one_to_many_wkt_no_holes(sdk: TravelTimeSdk):
    response = await sdk.time_map_fast_wkt_no_holes_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        travel_time=900,
        transportation=Transportation(type="public_transport"),
    )

    assert len(response.results) == 2


@pytest.mark.asyncio
async def test_many_to_one_wkt_no_holes(sdk: TravelTimeSdk):
    response = await sdk.time_map_fast_wkt_no_holes_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        travel_time=900,
        transportation=Transportation(type="public_transport"),
        one_to_many=False,
    )

    assert len(response.results) == 2



import math
from typing import Dict, Union, List, Optional

from traveltimepy.dto.requests import time_map_fast
from traveltimepy.dto.requests.distance_map import DistanceMapRequest
from traveltimepy.dto.requests.time_map_geojson import TimeMapRequestGeojson
from traveltimepy.dto.requests.time_map_wkt import TimeMapWKTRequest
from traveltimepy.errors import ApiError
from traveltimepy.proto import TimeFilterFastRequest_pb2

from traveltimepy.dto.common import (
    Location,
    Coordinates,
    FullRange,
    PolygonsFilter,
    Property,
    Range,
    LevelOfDetail,
    PropertyProto,
    TimeInfo,
    ArrivalTime,
    DepartureTime,
    Snapping,
)
from traveltimepy.dto.transportation import (
    PublicTransport,
    Driving,
    Ferry,
    Walking,
    Cycling,
    DrivingTrain,
    CyclingPublicTransport,
)
from traveltimepy.dto.requests.time_filter_fast import Transportation
from traveltimepy.dto.requests.postcodes_zones import ZonesProperty

from traveltimepy.dto.requests.postcodes import PostcodesRequest
from traveltimepy.dto.requests.routes import RoutesRequest
from traveltimepy.dto.requests.time_filter import TimeFilterRequest
from traveltimepy.dto.requests.time_filter_fast import TimeFilterFastRequest
from traveltimepy.dto.requests.time_map_fast import TimeMapFastRequest
from traveltimepy.dto.requests.time_map_fast_geojson import TimeMapFastGeojsonRequest
from traveltimepy.dto.requests.time_map_fast_wkt import TimeMapFastWKTRequest
from traveltimepy.dto.requests.time_filter_proto import ProtoTransportation
from traveltimepy.dto.requests.postcodes_zones import (
    PostcodesDistrictsRequest,
    PostcodesSectorsRequest,
)
from traveltimepy.dto.requests.time_map import TimeMapRequest

from traveltimepy.dto.requests import (
    distance_map,
    time_filter,
    time_filter_fast,
    postcodes,
    postcodes_zones,
    routes,
    time_map,
)


def create_time_filter(
    locations: List[Location],
    search_ids: Dict[str, List[str]],
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ],
    properties: Optional[List[Property]],
    time_info: TimeInfo,
    travel_time: int,
    range: Optional[FullRange],
    snapping: Optional[Snapping] = None,
) -> TimeFilterRequest:
    if properties is None:
        properties = [Property.TRAVEL_TIME]
    if isinstance(time_info, ArrivalTime):
        return TimeFilterRequest(
            locations=locations,
            arrival_searches=[
                time_filter.ArrivalSearch(
                    id=arrival_id,
                    arrival_location_id=arrival_id,
                    departure_location_ids=[
                        departure_id for departure_id in departure_ids
                    ],
                    arrival_time=time_info.value,
                    travel_time=travel_time,
                    transportation=transportation,
                    properties=properties,
                    range=range,
                    snapping=snapping,
                )
                for arrival_id, departure_ids in search_ids.items()
            ],
            departure_searches=[],
        )
    elif isinstance(time_info, DepartureTime):
        return TimeFilterRequest(
            locations=locations,
            departure_searches=[
                time_filter.DepartureSearch(
                    id=departure_id,
                    departure_location_id=departure_id,
                    arrival_location_ids=[arrival_id for arrival_id in arrival_ids],
                    departure_time=time_info.value,
                    travel_time=travel_time,
                    transportation=transportation,
                    properties=properties,
                    range=range,
                    snapping=snapping,
                )
                for departure_id, arrival_ids in search_ids.items()
            ],
            arrival_searches=[],
        )
    else:
        raise ApiError("arrival_time or departure_time should be specified")


def create_time_map_fast(
    coordinates: List[Coordinates],
    transportation: time_map_fast.Transportation,
    travel_time: int,
    level_of_detail: Optional[LevelOfDetail],
    snapping: Optional[Snapping],
    polygons_filter: Optional[PolygonsFilter],
    one_to_many: bool = True,
) -> TimeMapFastRequest:
    if one_to_many:
        return TimeMapFastRequest(
            arrival_searches=time_map_fast.ArrivalSearches(
                one_to_many=[
                    time_map_fast.Search(
                        id=f"Search {ind}",
                        coords=cur_coordinates,
                        transportation=transportation,
                        travel_time=travel_time,
                        arrival_time_period="weekday_morning",
                        level_of_detail=level_of_detail,
                        snapping=snapping,
                        polygons_filter=polygons_filter,
                    )
                    for ind, cur_coordinates in enumerate(coordinates)
                ],
                many_to_one=[],
            ),
        )
    else:
        return TimeMapFastRequest(
            arrival_searches=time_map_fast.ArrivalSearches(
                many_to_one=[
                    time_map_fast.Search(
                        id=f"Search {ind}",
                        coords=cur_coordinates,
                        transportation=transportation,
                        travel_time=travel_time,
                        arrival_time_period="weekday_morning",
                        level_of_detail=level_of_detail,
                        snapping=snapping,
                        polygons_filter=polygons_filter,
                    )
                    for ind, cur_coordinates in enumerate(coordinates)
                ],
                one_to_many=[],
            ),
        )


def create_time_map_fast_geojson(
    coordinates: List[Coordinates],
    transportation: time_map_fast.Transportation,
    travel_time: int,
    level_of_detail: Optional[LevelOfDetail],
    snapping: Optional[Snapping],
    polygons_filter: Optional[PolygonsFilter],
    one_to_many: bool = True,
) -> TimeMapFastGeojsonRequest:
    if one_to_many:
        return TimeMapFastGeojsonRequest(
            arrival_searches=time_map_fast.ArrivalSearches(
                one_to_many=[
                    time_map_fast.Search(
                        id=f"Search {ind}",
                        coords=cur_coordinates,
                        transportation=transportation,
                        travel_time=travel_time,
                        arrival_time_period="weekday_morning",
                        level_of_detail=level_of_detail,
                        snapping=snapping,
                        polygons_filter=polygons_filter,
                    )
                    for ind, cur_coordinates in enumerate(coordinates)
                ],
                many_to_one=[],
            ),
        )
    else:
        return TimeMapFastGeojsonRequest(
            arrival_searches=time_map_fast.ArrivalSearches(
                many_to_one=[
                    time_map_fast.Search(
                        id=f"Search {ind}",
                        coords=cur_coordinates,
                        transportation=transportation,
                        travel_time=travel_time,
                        arrival_time_period="weekday_morning",
                        level_of_detail=level_of_detail,
                        snapping=snapping,
                        polygons_filter=polygons_filter,
                    )
                    for ind, cur_coordinates in enumerate(coordinates)
                ],
                one_to_many=[],
            ),
        )


def create_time_map_fast_wkt(
    coordinates: List[Coordinates],
    transportation: time_map_fast.Transportation,
    travel_time: int,
    level_of_detail: Optional[LevelOfDetail],
    snapping: Optional[Snapping],
    polygons_filter: Optional[PolygonsFilter],
    one_to_many: bool = True,
) -> TimeMapFastWKTRequest:
    if one_to_many:
        return TimeMapFastWKTRequest(
            arrival_searches=time_map_fast.ArrivalSearches(
                one_to_many=[
                    time_map_fast.Search(
                        id=f"Search {ind}",
                        coords=cur_coordinates,
                        transportation=transportation,
                        travel_time=travel_time,
                        arrival_time_period="weekday_morning",
                        level_of_detail=level_of_detail,
                        snapping=snapping,
                        polygons_filter=polygons_filter,
                    )
                    for ind, cur_coordinates in enumerate(coordinates)
                ],
                many_to_one=[],
            ),
        )
    else:
        return TimeMapFastWKTRequest(
            arrival_searches=time_map_fast.ArrivalSearches(
                many_to_one=[
                    time_map_fast.Search(
                        id=f"Search {ind}",
                        coords=cur_coordinates,
                        transportation=transportation,
                        travel_time=travel_time,
                        arrival_time_period="weekday_morning",
                        level_of_detail=level_of_detail,
                        snapping=snapping,
                        polygons_filter=polygons_filter,
                    )
                    for ind, cur_coordinates in enumerate(coordinates)
                ],
                one_to_many=[],
            ),
        )


def create_time_filter_fast(
    locations: List[Location],
    search_ids: Dict[str, List[str]],
    transportation: Transportation,
    travel_time: int = 3600,
    properties: Optional[List[Property]] = None,
    one_to_many: bool = True,
    snapping: Optional[Snapping] = None,
) -> TimeFilterFastRequest:
    if properties is None:
        properties = [Property.TRAVEL_TIME]
    if one_to_many:
        return TimeFilterFastRequest(
            locations=locations,
            arrival_searches=time_filter_fast.ArrivalSearches(
                one_to_many=[
                    time_filter_fast.OneToMany(
                        id=departure_id,
                        departure_location_id=departure_id,
                        arrival_location_ids=arrival_ids,
                        transportation=transportation,
                        travel_time=travel_time,
                        arrival_time_period="weekday_morning",
                        properties=properties,
                        snapping=snapping,
                    )
                    for departure_id, arrival_ids in search_ids.items()
                ],
                many_to_one=[],
            ),
        )
    else:
        return TimeFilterFastRequest(
            locations=locations,
            arrival_searches=time_filter_fast.ArrivalSearches(
                many_to_one=[
                    time_filter_fast.ManyToOne(
                        id=arrival_id,
                        arrival_location_id=arrival_id,
                        departure_location_ids=departure_ids,
                        transportation=transportation,
                        travel_time=travel_time,
                        arrival_time_period="weekday_morning",
                        properties=properties,
                        snapping=snapping,
                    )
                    for arrival_id, departure_ids in search_ids.items()
                ],
                one_to_many=[],
            ),
        )


def create_postcodes(
    coordinates: List[Coordinates],
    time_info: TimeInfo,
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ],
    travel_time: int,
    properties: Optional[List[Property]],
    range: Optional[FullRange],
) -> PostcodesRequest:
    if properties is None:
        properties = [Property.TRAVEL_TIME]

    if isinstance(time_info, DepartureTime):
        return PostcodesRequest(
            departure_searches=[
                postcodes.DepartureSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=time_info.value,
                    transportation=transportation,
                    properties=properties,
                    range=range,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[],
        )
    elif isinstance(time_info, ArrivalTime):
        return PostcodesRequest(
            arrival_searches=[
                postcodes.ArrivalSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=time_info.value,
                    transportation=transportation,
                    properties=properties,
                    range=range,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[],
        )
    else:
        raise ApiError("arrival_time or departure_time should be specified")


def create_districts(
    coordinates: List[Coordinates],
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ],
    travel_time: int,
    time_info: TimeInfo,
    reachable_postcodes_threshold,
    properties: Optional[List[ZonesProperty]],
    range: Optional[FullRange],
) -> PostcodesDistrictsRequest:
    if properties is None:
        properties = [ZonesProperty.TRAVEL_TIME_ALL]

    if isinstance(time_info, ArrivalTime):
        return PostcodesDistrictsRequest(
            arrival_searches=[
                postcodes_zones.ArrivalSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=time_info.value,
                    reachable_postcodes_threshold=reachable_postcodes_threshold,
                    transportation=transportation,
                    properties=properties,
                    range=range,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[],
        )
    elif isinstance(time_info, DepartureTime):
        return PostcodesDistrictsRequest(
            departure_searches=[
                postcodes_zones.DepartureSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=time_info.value,
                    reachable_postcodes_threshold=reachable_postcodes_threshold,
                    transportation=transportation,
                    properties=properties,
                    range=range,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[],
        )
    else:
        raise ApiError("arrival_time or departure_time should be specified")


def create_sectors(
    coordinates: List[Coordinates],
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ],
    travel_time: int,
    time_info: TimeInfo,
    reachable_postcodes_threshold,
    properties: Optional[List[ZonesProperty]],
    range: Optional[FullRange],
) -> PostcodesSectorsRequest:
    if properties is None:
        properties = [ZonesProperty.TRAVEL_TIME_ALL]

    if isinstance(time_info, ArrivalTime):
        return PostcodesSectorsRequest(
            arrival_searches=[
                postcodes_zones.ArrivalSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=time_info.value,
                    reachable_postcodes_threshold=reachable_postcodes_threshold,
                    transportation=transportation,
                    properties=properties,
                    range=range,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[],
        )
    elif isinstance(time_info, DepartureTime):
        return PostcodesSectorsRequest(
            departure_searches=[
                postcodes_zones.DepartureSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=time_info.value,
                    reachable_postcodes_threshold=reachable_postcodes_threshold,
                    transportation=transportation,
                    properties=properties,
                    range=range,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[],
        )
    else:
        raise ApiError("arrival_time or departure_time should be specified")


def create_time_map(
    coordinates: List[Coordinates],
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ],
    travel_time: int,
    time_info: TimeInfo,
    search_range: Optional[Range],
    level_of_detail: Optional[LevelOfDetail],
    snapping: Optional[Snapping],
    polygons_filter: Optional[PolygonsFilter],
    remove_water_bodies: Optional[bool],
) -> TimeMapRequest:
    if isinstance(time_info, ArrivalTime):
        return TimeMapRequest(
            arrival_searches=[
                time_map.ArrivalSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=time_info.value,
                    transportation=transportation,
                    range=search_range,
                    level_of_detail=level_of_detail,
                    snapping=snapping,
                    polygons_filter=polygons_filter,
                    remove_water_bodies=remove_water_bodies,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[],
            unions=[],
            intersections=[],
        )
    elif isinstance(time_info, DepartureTime):
        return TimeMapRequest(
            departure_searches=[
                time_map.DepartureSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=time_info.value,
                    transportation=transportation,
                    range=search_range,
                    level_of_detail=level_of_detail,
                    snapping=snapping,
                    polygons_filter=polygons_filter,
                    remove_water_bodies=remove_water_bodies,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[],
            unions=[],
            intersections=[],
        )
    else:
        raise ApiError("arrival_time or departure_time should be specified")


def create_time_map_geojson(
    coordinates: List[Coordinates],
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ],
    travel_time: int,
    time_info: TimeInfo,
    search_range: Optional[Range],
    level_of_detail: Optional[LevelOfDetail],
    snapping: Optional[Snapping],
    polygons_filter: Optional[PolygonsFilter],
    remove_water_bodies: Optional[bool],
) -> TimeMapRequestGeojson:
    if isinstance(time_info, ArrivalTime):
        return TimeMapRequestGeojson(
            arrival_searches=[
                time_map.ArrivalSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=time_info.value,
                    transportation=transportation,
                    range=search_range,
                    level_of_detail=level_of_detail,
                    snapping=snapping,
                    polygons_filter=polygons_filter,
                    remove_water_bodies=remove_water_bodies,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[],
        )
    elif isinstance(time_info, DepartureTime):
        return TimeMapRequestGeojson(
            departure_searches=[
                time_map.DepartureSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=time_info.value,
                    transportation=transportation,
                    range=search_range,
                    level_of_detail=level_of_detail,
                    snapping=snapping,
                    polygons_filter=polygons_filter,
                    remove_water_bodies=remove_water_bodies,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[],
        )
    else:
        raise ApiError("arrival_time or departure_time should be specified")


def create_time_map_wkt(
    coordinates: List[Coordinates],
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ],
    travel_time: int,
    time_info: TimeInfo,
    search_range: Optional[Range],
    level_of_detail: Optional[LevelOfDetail],
    snapping: Optional[Snapping],
    polygons_filter: Optional[PolygonsFilter],
    remove_water_bodies: Optional[bool],
) -> TimeMapWKTRequest:
    if isinstance(time_info, ArrivalTime):
        return TimeMapWKTRequest(
            arrival_searches=[
                time_map.ArrivalSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=time_info.value,
                    transportation=transportation,
                    range=search_range,
                    level_of_detail=level_of_detail,
                    snapping=snapping,
                    polygons_filter=polygons_filter,
                    remove_water_bodies=remove_water_bodies,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[],
        )
    elif isinstance(time_info, DepartureTime):
        return TimeMapWKTRequest(
            departure_searches=[
                time_map.DepartureSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=time_info.value,
                    transportation=transportation,
                    range=search_range,
                    snapping=snapping,
                    polygons_filter=polygons_filter,
                    remove_water_bodies=remove_water_bodies,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[],
        )
    else:
        raise ApiError("arrival_time or departure_time should be specified")


def create_distance_map(
    coordinates: List[Coordinates],
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ],
    travel_distance: int,
    time_info: TimeInfo,
    level_of_detail: Optional[LevelOfDetail],
    snapping: Optional[Snapping],
    polygons_filter: Optional[PolygonsFilter],
    no_holes: Optional[bool],
) -> DistanceMapRequest:
    if isinstance(time_info, ArrivalTime):
        return DistanceMapRequest(
            arrival_searches=[
                distance_map.ArrivalSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_distance=travel_distance,
                    arrival_time=time_info.value,
                    transportation=transportation,
                    level_of_detail=level_of_detail,
                    snapping=snapping,
                    polygons_filter=polygons_filter,
                    no_holes=no_holes,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[],
            unions=[],
            intersections=[],
        )
    elif isinstance(time_info, DepartureTime):
        return DistanceMapRequest(
            departure_searches=[
                distance_map.DepartureSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_distance=travel_distance,
                    departure_time=time_info.value,
                    transportation=transportation,
                    level_of_detail=level_of_detail,
                    snapping=snapping,
                    polygons_filter=polygons_filter,
                    no_holes=no_holes,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[],
            unions=[],
            intersections=[],
        )
    else:
        raise ApiError("arrival_time or departure_time should be specified")


def create_intersection(
    coordinates: List[Coordinates],
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ],
    travel_time: int,
    time_info: TimeInfo,
    search_range: Optional[Range],
    level_of_detail: Optional[LevelOfDetail],
    snapping: Optional[Snapping],
    polygons_filter: Optional[PolygonsFilter],
    remove_water_bodies: Optional[bool],
) -> TimeMapRequest:
    if isinstance(time_info, ArrivalTime):
        return TimeMapRequest(
            arrival_searches=[
                time_map.ArrivalSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=time_info.value,
                    transportation=transportation,
                    range=search_range,
                    level_of_detail=level_of_detail,
                    snapping=snapping,
                    polygons_filter=polygons_filter,
                    remove_water_bodies=remove_water_bodies,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[],
            unions=[],
            intersections=[
                time_map.Intersection(
                    id="Intersection search",
                    search_ids=[f"Search {ind}" for ind, _ in enumerate(coordinates)],
                )
            ],
        )
    elif isinstance(time_info, DepartureTime):
        return TimeMapRequest(
            departure_searches=[
                time_map.DepartureSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=time_info.value,
                    transportation=transportation,
                    range=search_range,
                    level_of_detail=level_of_detail,
                    snapping=snapping,
                    polygons_filter=polygons_filter,
                    remove_water_bodies=remove_water_bodies,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[],
            unions=[],
            intersections=[
                time_map.Intersection(
                    id="Intersection search",
                    search_ids=[f"Search {ind}" for ind, _ in enumerate(coordinates)],
                )
            ],
        )
    else:
        raise ApiError("arrival_time or departure_time should be specified")


def create_union(
    coordinates: List[Coordinates],
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ],
    travel_time: int,
    time_info: TimeInfo,
    search_range: Optional[Range],
    level_of_detail: Optional[LevelOfDetail],
    snapping: Optional[Snapping],
    polygons_filter: Optional[PolygonsFilter],
    remove_water_bodies: Optional[bool],
) -> TimeMapRequest:
    if isinstance(time_info, ArrivalTime):
        return TimeMapRequest(
            arrival_searches=[
                time_map.ArrivalSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=time_info.value,
                    transportation=transportation,
                    range=search_range,
                    level_of_detail=level_of_detail,
                    snapping=snapping,
                    polygons_filter=polygons_filter,
                    remove_water_bodies=remove_water_bodies,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[],
            unions=[
                time_map.Union(
                    id="Union search",
                    search_ids=[f"Search {ind}" for ind, _ in enumerate(coordinates)],
                )
            ],
            intersections=[],
        )
    elif isinstance(time_info, DepartureTime):
        return TimeMapRequest(
            departure_searches=[
                time_map.DepartureSearch(
                    id=f"Search {ind}",
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=time_info.value,
                    transportation=transportation,
                    range=search_range,
                    level_of_detail=level_of_detail,
                    snapping=snapping,
                    polygons_filter=polygons_filter,
                    remove_water_bodies=remove_water_bodies,
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[],
            unions=[
                time_map.Union(
                    id="Union search",
                    search_ids=[f"Search {ind}" for ind, _ in enumerate(coordinates)],
                )
            ],
            intersections=[],
        )
    else:
        raise ApiError("arrival_time or departure_time should be specified")


def create_routes(
    locations: List[Location],
    search_ids: Dict[str, List[str]],
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ],
    time_info: TimeInfo,
    properties: Optional[List[Property]],
    range: Optional[FullRange],
    snapping: Optional[Snapping],
) -> RoutesRequest:
    if properties is None:
        properties = [Property.TRAVEL_TIME]
    if isinstance(time_info, ArrivalTime):
        return RoutesRequest(
            locations=locations,
            arrival_searches=[
                routes.ArrivalSearch(
                    id=arrival_id,
                    arrival_location_id=arrival_id,
                    departure_location_ids=[
                        departure_id for departure_id in departure_ids
                    ],
                    arrival_time=time_info.value,
                    transportation=transportation,
                    properties=properties,
                    range=range,
                    snapping=snapping,
                )
                for arrival_id, departure_ids in search_ids.items()
            ],
            departure_searches=[],
        )
    elif isinstance(time_info, DepartureTime):
        return RoutesRequest(
            locations=locations,
            departure_searches=[
                routes.DepartureSearch(
                    id=departure_id,
                    departure_location_id=departure_id,
                    arrival_location_ids=[arrival_id for arrival_id in arrival_ids],
                    departure_time=time_info.value,
                    transportation=transportation,
                    properties=properties,
                    range=range,
                    snapping=snapping,
                )
                for departure_id, arrival_ids in search_ids.items()
            ],
            arrival_searches=[],
        )
    else:
        raise ApiError("arrival_time or departure_time should be specified")


def create_proto_request(
    origin: Coordinates,
    destinations: List[Coordinates],
    transportation: ProtoTransportation,
    properties: Optional[List[PropertyProto]],
    travel_time: int,
    one_to_many: bool = True,
) -> TimeFilterFastRequest_pb2.TimeFilterFastRequest:  # type: ignore
    request = TimeFilterFastRequest_pb2.TimeFilterFastRequest()  # type: ignore

    if one_to_many:
        request.oneToManyRequest.departureLocation.lat = origin.lat
        request.oneToManyRequest.departureLocation.lng = origin.lng

        request.oneToManyRequest.transportation.type = transportation.value.code
        request.oneToManyRequest.travelTime = travel_time
        request.oneToManyRequest.arrivalTimePeriod = (
            TimeFilterFastRequest_pb2.TimePeriod.WEEKDAY_MORNING  # type: ignore
        )
        if properties is not None:
            request.oneToManyRequest.properties.extend(properties)

        mult = math.pow(10, 5)
        for destination in destinations:
            lat_delta = round((destination.lat - origin.lat) * mult)
            lng_delta = round((destination.lng - origin.lng) * mult)
            request.oneToManyRequest.locationDeltas.extend([lat_delta, lng_delta])
    else:
        request.manyToOneRequest.arrivalLocation.lat = origin.lat
        request.manyToOneRequest.arrivalLocation.lng = origin.lng

        request.manyToOneRequest.transportation.type = transportation.value.code
        request.manyToOneRequest.travelTime = travel_time
        request.manyToOneRequest.arrivalTimePeriod = (
            TimeFilterFastRequest_pb2.TimePeriod.WEEKDAY_MORNING  # type: ignore
        )
        if properties is not None:
            request.manyToOneRequest.properties.extend(properties)

        mult = math.pow(10, 5)
        for destination in destinations:
            lat_delta = round((destination.lat - origin.lat) * mult)
            lng_delta = round((destination.lng - origin.lng) * mult)
            request.manyToOneRequest.locationDeltas.extend([lat_delta, lng_delta])

    return request



"""Module with WKT wrapper for shapely"""

from traveltimepy.wkt.parsing import (
    parse_wkt,
)

from traveltimepy.wkt.geometries import (
    WKTObject,
    PointModel,
    LineStringModel,
    PolygonModel,
    MultiLineStringModel,
    MultiPointModel,
    MultiPolygonModel,
    GeometryType,
)

__all__ = [
    "WKTObject",
    "PointModel",
    "LineStringModel",
    "PolygonModel",
    "MultiPointModel",
    "MultiLineStringModel",
    "MultiPolygonModel",
    "GeometryType",
    "parse_wkt",
]



import itertools
from typing import List, TypeVar, Tuple, Optional

T = TypeVar("T")
R = TypeVar("R")


def sliding(values: List[T], window_size: int) -> List[List[T]]:
    return [values[i : i + window_size] for i in range(0, len(values), window_size)]


def split(
    left: List[T], right: List[R], window_size: int
) -> List[Tuple[List[T], List[R]]]:
    return list(
        itertools.zip_longest(
            sliding(left, window_size), sliding(right, window_size), fillvalue=[]
        )
    )


def join_opt(values: Optional[List[str]], sep: str) -> Optional[str]:
    return sep.join(values) if values is not None and len(values) != 0 else None


def flatten(list_of_lists: List[List[T]]) -> List[T]:
    return list(itertools.chain.from_iterable(list_of_lists))



from typing import Dict

from aiohttp import (
    ClientSession,
    ClientResponse,
    BasicAuth,
    TCPConnector,
    ClientTimeout,
)

from traveltimepy.proto import TimeFilterFastResponse_pb2, TimeFilterFastRequest_pb2
from traveltimepy.dto.responses.time_filter_proto import TimeFilterProtoResponse
from traveltimepy.errors import ApiError


async def send_proto_async(
    url: str,
    headers: Dict[str, str],
    data: TimeFilterFastRequest_pb2.TimeFilterFastRequest,  # type: ignore
    app_id: str,
    api_key: str,
    timeout: int,
) -> TimeFilterProtoResponse:
    async with ClientSession(
        timeout=ClientTimeout(total=timeout), connector=TCPConnector(ssl=False)
    ) as session:
        async with session.post(
            url=url,
            headers=headers,
            data=data.SerializeToString(),
            auth=BasicAuth(app_id, api_key),
        ) as resp:
            return await _process_response(resp)


async def _process_response(response: ClientResponse) -> TimeFilterProtoResponse:
    content = await response.read()
    if response.status != 200:
        msg = (
            f"Travel Time API proto request failed with error code: {response.status}\n"
        )
        raise ApiError(msg)
    else:
        response_body = TimeFilterFastResponse_pb2.TimeFilterFastResponse()  # type: ignore
        response_body.ParseFromString(content)
        return TimeFilterProtoResponse(
            travel_times=response_body.properties.travelTimes[:],
            distances=response_body.properties.distances[:],
        )



import pytest
from datetime import datetime

from traveltimepy import Coordinates, Driving, LevelOfDetail, TravelTimeSdk


@pytest.mark.asyncio
async def test_departures(sdk: TravelTimeSdk):
    results = await sdk.distance_map_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        departure_time=datetime.now(),
        travel_distance=900,
        transportation=Driving(),
        level_of_detail=LevelOfDetail(scale_type="simple", level="lowest"),
    )
    assert len(results) == 2


@pytest.mark.asyncio
async def test_arrivals(sdk: TravelTimeSdk):
    results = await sdk.distance_map_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        arrival_time=datetime.now(),
        travel_distance=900,
        transportation=Driving(),
        level_of_detail=LevelOfDetail(scale_type="simple", level="lowest"),
    )
    assert len(results) == 2



import pytest  # noqa

from traveltimepy import Coordinates
from traveltimepy.wkt import LineStringModel, PointModel

mock_point1 = PointModel(coordinates=Coordinates(lat=10, lng=20))
mock_point2 = PointModel(coordinates=Coordinates(lat=30, lng=40))


def test_linestring_minimum_coordinates_valid():
    # Create LineStringModel with two valid coordinates
    linestring = LineStringModel(coordinates=[mock_point1, mock_point2])
    assert len(linestring.coordinates) == 2


def test_linestring_minimum_coordinates_invalid():
    # Test LineStringModel with less than two coordinates
    with pytest.raises(ValueError) as excinfo:
        LineStringModel(coordinates=[mock_point1])
    assert "LineString must have at least 2 coordinates." in str(excinfo.value)



from functools import singledispatch
from typing import Union

from shapely import wkt, GEOSException
from shapely.geometry import (
    Point,
    LineString,
    Polygon,
    MultiPoint,
    MultiLineString,
    MultiPolygon,
)
from shapely.geometry.base import BaseGeometry

from traveltimepy import Coordinates
from traveltimepy.wkt.geometries import (
    PointModel,
    LineStringModel,
    PolygonModel,
    MultiPointModel,
    MultiLineStringModel,
    MultiPolygonModel,
)
from traveltimepy.wkt.constants import (
    SUPPORTED_GEOMETRY_TYPES,
)

from traveltimepy.wkt.error import (
    InvalidWKTStringError,
    NullGeometryError,
    InvalidFunctionError,
    InvalidGeometryTypeError,
)


def _check_empty(geometry):
    if geometry.is_empty or geometry is None:
        raise NullGeometryError()


@singledispatch
def _parse_geometry(geometry: BaseGeometry):
    raise InvalidFunctionError(geometry)


@_parse_geometry.register
def _parse_point(geometry: Point) -> PointModel:
    return PointModel(coordinates=Coordinates(lat=geometry.y, lng=geometry.x))


@_parse_geometry.register
def _parse_line_string(geometry: LineString) -> LineStringModel:
    point_models = [
        PointModel(coordinates=Coordinates(lat=lat, lng=lng))
        for lat, lng in geometry.coords
    ]
    return LineStringModel(coordinates=point_models)


@_parse_geometry.register
def _parse_polygon(geometry: Polygon) -> PolygonModel:
    exterior_points = [
        PointModel(coordinates=Coordinates(lat=lat, lng=lng))
        for lat, lng in geometry.exterior.coords
    ]
    interiors_points = [
        [
            PointModel(coordinates=Coordinates(lat=lat, lng=lng))
            for lat, lng in interior.coords
        ]
        for interior in list(geometry.interiors)
    ]
    exterior_line = LineStringModel(coordinates=exterior_points)
    interior_lines = [
        LineStringModel(coordinates=points) for points in interiors_points
    ]
    return PolygonModel(exterior=exterior_line, interiors=interior_lines)


@_parse_geometry.register
def _parse_multi_point(geometry: MultiPoint) -> MultiPointModel:
    point_models = [
        PointModel(coordinates=Coordinates(lat=point.y, lng=point.x))
        for point in geometry.geoms
    ]
    return MultiPointModel(coordinates=point_models)


@_parse_geometry.register
def _parse_multi_line_string(geometry: MultiLineString) -> MultiLineStringModel:
    line_strings = [
        LineStringModel(
            coordinates=[
                PointModel(
                    coordinates=Coordinates(lat=point[0], lng=point[1]),
                )
                for point in linestring.coords
            ],
        )
        for linestring in geometry.geoms
    ]
    return MultiLineStringModel(coordinates=line_strings)


@_parse_geometry.register
def _parse_multi_polygon(geometry: MultiPolygon) -> MultiPolygonModel:
    polygons = [
        PolygonModel(
            exterior=LineStringModel(
                coordinates=[
                    PointModel(
                        coordinates=Coordinates(lat=point[0], lng=point[1]),
                    )
                    for point in polygon.exterior.coords
                ],
            ),
            interiors=[
                LineStringModel(
                    coordinates=[
                        PointModel(
                            coordinates=Coordinates(lat=point[0], lng=point[1]),
                        )
                        for point in interior.coords
                    ],
                )
                for interior in list(polygon.interiors)
            ],
        )
        for polygon in geometry.geoms
    ]
    return MultiPolygonModel(coordinates=polygons)


def parse_wkt(
    wkt_str: str,
) -> Union[
    PointModel,
    LineStringModel,
    PolygonModel,
    MultiPointModel,
    MultiLineStringModel,
    MultiPolygonModel,
]:
    try:
        geometry = wkt.loads(wkt_str)
    except GEOSException:
        raise InvalidWKTStringError(wkt_str)

    _check_empty(geometry)

    if type(geometry) not in SUPPORTED_GEOMETRY_TYPES:
        raise InvalidGeometryTypeError(geometry)

    return _parse_geometry(geometry)



from typing import List

from pydantic.main import BaseModel

from traveltimepy.dto.common import Coordinates


class Shape(BaseModel):
    shell: List[Coordinates]
    holes: List[List[Coordinates]]


class TimeMapResult(BaseModel):
    search_id: str
    shapes: List[Shape]


class TimeMapResponse(BaseModel):
    results: List[TimeMapResult]



from typing import List
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.requests.time_map_fast import ArrivalSearches
from traveltimepy.dto.responses.time_map_wkt import TimeMapWKTResponse
from traveltimepy.itertools import split, flatten


class TimeMapFastWKTRequest(TravelTimeRequest[TimeMapWKTResponse]):
    arrival_searches: ArrivalSearches

    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        return [
            TimeMapFastWKTRequest(
                arrival_searches=ArrivalSearches(
                    one_to_many=one_to_many, many_to_one=many_to_one
                ),
            )
            for one_to_many, many_to_one in split(
                self.arrival_searches.one_to_many,
                self.arrival_searches.many_to_one,
                window_size,
            )
        ]

    def merge(self, responses: List[TimeMapWKTResponse]) -> TimeMapWKTResponse:
        return TimeMapWKTResponse(
            results=sorted(
                flatten([response.results for response in responses]),
                key=lambda res: res.search_id,
            )
        )



from typing import List, Optional
from typing_extensions import Literal

from pydantic import BaseModel

from traveltimepy.dto.common import Location, Property, Snapping
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.responses.time_filter_fast import TimeFilterFastResponse
from traveltimepy.itertools import split, flatten


class Transportation(BaseModel):
    type: Literal[
        "public_transport",
        "driving",
        "cycling",
        "walking",
        "walking+ferry",
        "cycling+ferry",
        "driving+ferry",
        "driving+public_transport",
    ]


class OneToMany(BaseModel):
    id: str
    departure_location_id: str
    arrival_location_ids: List[str]
    transportation: Transportation
    travel_time: int
    arrival_time_period: str
    properties: List[Property]
    snapping: Optional[Snapping] = None


class ManyToOne(BaseModel):
    id: str
    arrival_location_id: str
    departure_location_ids: List[str]
    transportation: Transportation
    travel_time: int
    arrival_time_period: str
    properties: List[Property]
    snapping: Optional[Snapping] = None


class ArrivalSearches(BaseModel):
    many_to_one: List[ManyToOne]
    one_to_many: List[OneToMany]


class TimeFilterFastRequest(TravelTimeRequest[TimeFilterFastResponse]):
    locations: List[Location]
    arrival_searches: ArrivalSearches

    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        return [
            TimeFilterFastRequest(
                locations=self.locations,
                arrival_searches=ArrivalSearches(
                    one_to_many=one_to_many, many_to_one=many_to_one
                ),
            )
            for one_to_many, many_to_one in split(
                self.arrival_searches.one_to_many,
                self.arrival_searches.many_to_one,
                window_size,
            )
        ]

    def merge(self, responses: List[TimeFilterFastResponse]) -> TimeFilterFastResponse:
        return TimeFilterFastResponse(
            results=flatten([response.results for response in responses])
        )



import asyncio
import json
from dataclasses import dataclass
from typing import TypeVar, Type, Dict, Optional

from aiohttp import ClientSession, ClientResponse, TCPConnector, ClientTimeout
from pydantic import BaseModel
from traveltimepy.dto.requests.request import TravelTimeRequest

from traveltimepy.dto.responses.error import ResponseError
from traveltimepy.errors import ApiError
from aiohttp_retry import RetryClient, ExponentialRetry
from aiolimiter import AsyncLimiter

T = TypeVar("T", bound=BaseModel)
DEFAULT_SPLIT_SIZE = 10


@dataclass
class SdkParams:
    host: str
    proto_host: str
    limit_per_host: int
    rate_limit: int
    time_window: int
    retry_attempts: int
    timeout: int


async def send_post_request_async(
    client: RetryClient,
    response_class: Type[T],
    url: str,
    headers: Dict[str, str],
    request: TravelTimeRequest,
    rate_limit: AsyncLimiter,
) -> T:
    async with rate_limit:
        async with client.post(
            url=url, headers=headers, data=request.model_dump_json()
        ) as resp:
            return await _process_response(response_class, resp)


async def send_post_async(
    response_class: Type[T],
    path: str,
    headers: Dict[str, str],
    request: TravelTimeRequest,
    sdk_params: SdkParams,
) -> T:
    window_size = _window_size(sdk_params.rate_limit)
    async with ClientSession(
        timeout=ClientTimeout(total=sdk_params.timeout),
        connector=TCPConnector(ssl=False, limit_per_host=sdk_params.limit_per_host),
    ) as session:
        retry_options = ExponentialRetry(attempts=sdk_params.retry_attempts)
        async with RetryClient(
            client_session=session, retry_options=retry_options
        ) as client:
            rate_limit = AsyncLimiter(
                sdk_params.rate_limit // window_size, sdk_params.time_window
            )
            tasks = [
                send_post_request_async(
                    client,
                    response_class,
                    f"https://{sdk_params.host}/v4/{path}",
                    headers,
                    part,
                    rate_limit,
                )
                for part in request.split_searches(window_size)
            ]
            responses = await asyncio.gather(*tasks)
            return request.merge(responses)


def _window_size(rate_limit: int):
    if rate_limit >= DEFAULT_SPLIT_SIZE:
        return DEFAULT_SPLIT_SIZE
    else:
        return rate_limit


async def send_get_async(
    response_class: Type[T],
    path: str,
    headers: Dict[str, str],
    sdk_params: SdkParams,
    params: Optional[Dict[str, str]],
) -> T:
    async with ClientSession(
        timeout=ClientTimeout(total=sdk_params.timeout),
        connector=TCPConnector(ssl=False),
    ) as session:
        retry_options = ExponentialRetry(attempts=sdk_params.retry_attempts)
        async with RetryClient(
            client_session=session, retry_options=retry_options
        ) as client:
            async with client.get(
                url=f"https://{sdk_params.host}/v4/{path}",
                headers=headers,
                params=params,
            ) as resp:
                return await _process_response(response_class, resp)


async def _process_response(response_class: Type[T], response: ClientResponse) -> T:
    text = await response.text()
    json_data = json.loads(text)
    if response.status != 200:
        parsed = ResponseError.model_validate_json(json.dumps(json_data))
        msg = (
            f"Travel Time API request failed: {parsed.description}\n"
            f"Error code: {parsed.error_code}\n"
            f"Additional info: {parsed.additional_info}\n"
            f"<{parsed.documentation_link}>\n"
        )
        raise ApiError(msg)
    else:
        return response_class.model_validate(json_data)



from typing import List

from traveltimepy.dto.common import Location
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.responses.supported_locations import SupportedLocationsResponse
from traveltimepy.itertools import flatten


class SupportedLocationsRequest(TravelTimeRequest[SupportedLocationsResponse]):
    locations: List[Location]

    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        return [SupportedLocationsRequest(locations=self.locations)]

    def merge(
        self, responses: List[SupportedLocationsResponse]
    ) -> SupportedLocationsResponse:
        return SupportedLocationsResponse(
            locations=flatten([response.locations for response in responses]),
            unsupported_locations=flatten(
                [response.unsupported_locations for response in responses]
            ),
        )



from typing import List

from pydantic import BaseModel


class SupportedLocation(BaseModel):
    id: str
    map_name: str
    additional_map_names: List[str]


class SupportedLocationsResponse(BaseModel):
    locations: List[SupportedLocation]
    unsupported_locations: List[str]



from typing import Generic, List, Union, Dict, Any, TypeVar, Optional
from pydantic import field_validator, BaseModel, Field

from traveltimepy.wkt import WKTObject, parse_wkt
from traveltimepy.wkt.helper import print_indented

Props = TypeVar("Props", bound=Union[Dict[str, Any], BaseModel])


class TimeMapWKTResult(BaseModel, Generic[Props]):
    search_id: str
    shape: WKTObject
    properties: Optional[Props] = Field(None)

    @field_validator("shape", mode="before")
    @classmethod
    def transform_shape(cls, shape: str) -> WKTObject:
        return parse_wkt(shape)

    def pretty_print(self, indent_level=0):
        print_indented(f"SEARCH ID: {self.search_id}", indent_level)
        self.shape.pretty_print(indent_level)
        print_indented(f"PROPERTIES: {self.properties}", indent_level)


class TimeMapWKTResponse(BaseModel):
    results: List[TimeMapWKTResult]

    def pretty_print(self, indent_level=0):
        print_indented("TIME-MAP WKT RESPONSE:", indent_level)
        for result in self.results:
            result.pretty_print(indent_level + 1)
            print()
